#!/usr/bin/env python
"""
Tests for c64img.hires module
This is part of the c64img project <https://bitbucket.org/gryf/image2c64>

Author: Roman 'gryf' Dobosz <gryf73@gmail.com>
Licence: BSD
"""
import os
from unittest import TestCase, main
from tempfile import mkstemp

from PIL import Image

from c64img import hires


Image.Image.show = lambda x: None


def get_image_path(image_name):
    return os.path.join(os.path.dirname(__file__), "test_images", image_name)


CLASH_H = get_image_path("clash.hires.png")
CLASH_M = get_image_path("clash.multi.png")
CLASH_M2 = get_image_path("clash.multi.160.png")
COLORS_256 = get_image_path("colors.256.png")
PAL_PEPTO = get_image_path("colors.pepto.png")
PAL_TIMANTHES = get_image_path("colors.timanthes.png")
PAL_UNKNOWN = get_image_path("colors.unknown.png")
PAL_VICE = get_image_path("colors.vice.png")
CROP_BOTH = get_image_path("crop.251x187.png")
CROP_H = get_image_path("crop.319x200.png")
CROP_V = get_image_path("crop.320x199.png")
COLORS_1 = get_image_path("hires_1c.png")
COLORS_2 = get_image_path("hires_2c.png")
COLORS_256_U16 = get_image_path("hires_256_16p.png")  # 256 defined, uniq 16
HIRES = get_image_path("hires.png")
MULTI = get_image_path("multi.160x200.png")
MULTI_320 = get_image_path("multi.320x200.png")


class LogMock:
    """Mock logger class"""

    def warning(*args, **kwargs):
        return

    def debug(*args, **kwargs):
        return


class Interceptor:
    """
    Interceptor class for function call detection
    """
    def __init__(self):
        """
        Init.
        """
        self.call = 0
        self.repeat = False

    def __call__(self, dummy1, dummy2=False):
        """
        Call attribute is increased every time, instance is called
        """
        self.call += 1
        return self.repeat


class TestHiresChar(TestCase):
    """
    Test hires Char class
    """
    def setUp(self):
        """set up"""
        self.log = LogMock()

    def test_get_binary_data(self):
        """
        Test get_binary_data method
        """
        char = hires.HiresChar(self.log)
        char.pixels = {(0, 0): 1, (0, 1): 0, (0, 2): 1, (0, 3): 0,
                       (0, 4): 0, (0, 5): 1, (0, 6): 0, (0, 7): 1}
        char._analyze_color_map()
        result = char.get_binary_data()
        self.assertEqual(result['bitmap'], [0b10100101])
        self.assertEqual(result['screen-ram'], 0x10)

        # last pixel with the clash - should fall back to background color
        char.pixels = {(0, 0): 1, (0, 1): 0, (0, 2): 1, (0, 3): 0,
                       (0, 4): 0, (0, 5): 1, (0, 6): 0, (0, 7): 1,
                       (1, 0): 1, (1, 1): 0, (1, 2): 1, (1, 3): 0,
                       (1, 4): 0, (1, 5): 1, (1, 6): 0, (1, 7): 1,
                       (2, 0): 1, (2, 1): 0, (2, 2): 1, (2, 3): 0,
                       (2, 4): 0, (2, 5): 1, (2, 6): 0, (2, 7): 1,
                       (3, 0): 1, (3, 1): 0, (3, 2): 1, (3, 3): 0,
                       (3, 4): 0, (3, 5): 1, (3, 6): 0, (3, 7): 1,
                       (4, 0): 1, (4, 1): 0, (4, 2): 1, (4, 3): 0,
                       (4, 4): 0, (4, 5): 1, (4, 6): 0, (4, 7): 1,
                       (5, 0): 1, (5, 1): 0, (5, 2): 1, (5, 3): 0,
                       (5, 4): 0, (5, 5): 1, (5, 6): 0, (5, 7): 1,
                       (6, 0): 1, (6, 1): 0, (6, 2): 1, (6, 3): 0,
                       (6, 4): 0, (6, 5): 1, (6, 6): 0, (6, 7): 1,
                       (7, 0): 1, (7, 1): 0, (7, 2): 1, (7, 3): 0,
                       (7, 4): 0, (7, 5): 1, (7, 6): 0, (7, 7): 2}
        char._fix_clash = True
        char._analyze_color_map()
        result = char.get_binary_data()
        self.assertEqual(result['bitmap'], [165, 165, 165, 165, 165, 165, 165,
                                            0b10100100])
        self.assertEqual(result['screen-ram'], 0x10)

    def test__compare_colors_with_prev_char(self):
        """
        Test _compare_colors_with_prev_char method. This method is responsible
        for creating new mapping for multicolor bit pairs in conjunction with
        corresponding colors.
        """
        char = hires.HiresChar(self.log)
        colors = {8: 50, 9: 14}

        # The default case. No previous picture stored, no colors to compare
        # with. None colors were stored.
        self.assertTrue(char._compare_colors_with_prev_char(colors))
        self.assertEqual(char.pixel_state, {0: False, 1: False})
        self.assertEqual(char.colors, {})

        # So it needs to rerun, colors are recognized and remembered.
        self.assertFalse(char._compare_colors_with_prev_char(colors, True))
        self.assertEqual(char.pixel_state, {0: True, 1: True})
        self.assertEqual(char.colors, {8: 0, 9: 1})

        # 1. Ideal case. Colors for previous and current character in char
        # boundary are the same. No need to rerun checks.
        char = hires.HiresChar(self.log)
        prev = hires.HiresChar(self.log)
        prev.colors = {8: 0, 9: 1}
        char.prev = prev
        self.assertFalse(char._compare_colors_with_prev_char(colors))
        self.assertEqual(char.pixel_state, {0: True, 1: True})
        self.assertEqual(char.colors, {8: 0, 9: 1})

        # 2. Mixed colors/pixel pairs. Color indices are matching fine. Colors
        # and pairs from previous character should be propagated into current
        # char.
        char = hires.HiresChar(self.log)
        prev = hires.HiresChar(self.log)
        prev.colors = {9: 1, 8: 0}
        char.prev = prev
        self.assertFalse(char._compare_colors_with_prev_char(colors))
        self.assertEqual(char.pixel_state, {0: True, 1: True})
        self.assertEqual(char.colors, {9: 1, 8: 0})

        # 3. Mixed colors/pixel pairs. One color index differ. Tho other color
        # from previous character should be propagated into current char,
        # the mismatch color should be replaced by current one.
        char = hires.HiresChar(self.log)
        prev = hires.HiresChar(self.log)
        prev.colors = {5: 1, 9: 0}
        char.prev = prev
        self.assertTrue(char._compare_colors_with_prev_char(colors))
        self.assertEqual(char.pixel_state, {0: True, 1: False})
        self.assertEqual(char.colors, {9: 0})
        self.assertFalse(char._compare_colors_with_prev_char(colors, True))
        self.assertEqual(char.pixel_state, {0: True, 1: True})
        self.assertEqual(char.colors, {9: 0, 8: 1})

        # 5. Worst case scenario. None of the colors from previous char
        # matches. Get the current colors.
        char = hires.HiresChar(self.log)
        prev = hires.HiresChar(self.log)
        prev.colors = {5: 1, 6: 0}
        char.prev = prev
        self.assertTrue(char._compare_colors_with_prev_char(colors))
        self.assertEqual(char.pixel_state, {0: False, 1: False})
        self.assertEqual(char.colors, {})
        self.assertFalse(char._compare_colors_with_prev_char(colors, True))
        self.assertEqual(char.pixel_state, {0: True, 1: True})
        self.assertEqual(char.colors, {9: 1, 8: 0})

    def test__fix_color_clash(self):
        """Test for repait color clash"""
        char = hires.HiresChar(self.log, prev=None, fix_clash=True)
        char.pixels = {(0, 0): 2, (0, 1): 1, (0, 2): 1, (0, 3): 7,
                       (0, 4): 2, (0, 5): 2, (0, 6): 3, (0, 7): 4}
        char._analyze_color_map()
        self.assertEqual(char.pixels,
                         {(0, 0): 2, (0, 1): 1, (0, 2): 1, (0, 3): 1,
                          (0, 4): 2, (0, 5): 2, (0, 6): 2, (0, 7): 2})


class TestHires(TestCase):
    """
    Tests for hires conversion
    """

    def test__check_dimensions(self):
        """
        Test of dimensions of the input images
        """
        for fname in (CROP_V, CROP_H, CROP_BOTH, CLASH_M2, MULTI, PAL_PEPTO,
                      PAL_TIMANTHES, PAL_UNKNOWN, PAL_VICE):
            obj = hires.HiresConverter(fname)
            obj.log.error = lambda *x: None  # suppress log
            obj._load()
            self.assertFalse(obj._check_dimensions())

        for fname in (CLASH_H, CLASH_M, COLORS_256, COLORS_1, COLORS_2,
                      COLORS_256_U16, HIRES, MULTI_320):
            obj = hires.HiresConverter(fname)
            obj._load()
            self.assertTrue(obj._check_dimensions())

    def test__fill_memory(self):
        """
        Test for _fill_memory method
        """
        obj = hires.HiresConverter(HIRES)
        obj._load()
        hist = obj._src_image.histogram()
        obj._colors_check(hist)
        obj._find_best_palette_map()
        obj._find_most_freq_color(hist)
        self.assertEqual(obj._fill_memory(), True)

        obj = hires.HiresConverter(COLORS_1)
        obj._load()
        hist = obj._src_image.histogram()
        obj.log.warn = lambda x, y: None  # suppress log
        obj._colors_check(hist)
        obj._find_best_palette_map()
        obj._find_most_freq_color(hist)
        self.assertEqual(obj._fill_memory(), True)

        obj = hires.HiresConverter(CLASH_H)
        obj._load()
        hist = obj._src_image.histogram()
        obj._colors_check(hist)
        obj._find_best_palette_map()
        obj._find_most_freq_color(hist)
        self.assertEqual(obj._fill_memory(), False)

    def test__get_displayer(self):
        """
        Test for _get_displayer method
        """
        obj = hires.HiresConverter(HIRES)
        border = 0
        # displayer routine is constant. On index 19 there is a border color.
        self.assertEqual(obj._get_displayer()[19], border)

        border = 10
        obj.set_border_color(border)
        self.assertEqual(obj._get_displayer()[19], border)

    def test_save(self):
        """
        Test for save method (and all _save... submethods)
        """
        fdesc, fname = mkstemp(suffix=".prg")
        os.close(fdesc)

        obj = hires.HiresConverter(HIRES)
        obj.log.warning = lambda *x: None  # suppress log
        self.assertEqual(obj.save(fname, "prg"), 0)

        self.assertTrue(os.path.exists(fname))
        self.assertEqual(os.stat(fname).st_size, 14145)
        os.unlink(fname)

        fdesc, fname = mkstemp(suffix=".prg")
        os.close(fdesc)

        obj = hires.HiresConverter(HIRES)
        obj.log.warning = lambda *x: None  # suppress log
        self.assertEqual(obj.save(fname, "art-studio-hires"), 0)

        self.assertTrue(os.path.exists(fname))
        self.assertEqual(os.stat(fname).st_size, 9009)
        os.unlink(fname)

        obj = hires.HiresConverter(CLASH_H)
        self.assertEqual(obj.save(fname, "art-studio-hires"), 1)

        obj = hires.HiresConverter(HIRES)
        self.assertEqual(obj.save(fname, "raw"), 0)

        self.assertTrue(os.path.exists(fname + "_bitmap.raw"))
        self.assertTrue(os.path.exists(fname + "_screen.raw"))

        self.assertEqual(os.stat(fname + "_bitmap.raw").st_size, 8000)
        self.assertEqual(os.stat(fname + "_screen.raw").st_size, 1000)
        os.unlink(fname + "_bitmap.raw")
        os.unlink(fname + "_screen.raw")


if __name__ == "__main__":
    main()



"""
image2c64 converts virtually any fullscreen image supported by Pillow to C64
hires or multicolor formats. Best results are achived with filetypes PNG or
GIF.
"""
import argparse
import os
import sys

from c64img import __version__ as ver
from c64img.hires import HiresConverter
from c64img.multi import MultiConverter
from c64img.path import get_modified_fname


def convert(arguments, converter_class):
    """
    Convert pictures
    """
    last = conv = None
    exit_code = 0
    for fname in arguments.filename:
        if conv:
            last = conv

        conv = converter_class(fname, arguments.errors)

        if last:
            conv.prev_chars = last.chars

        if arguments.border is not None:
            conv.set_border_color(arguments.border)

        # note, that for hires pictures it doesn't make sense, and will be
        # ignored.
        if arguments.background is not None:
            conv.set_bg_color(arguments.background)

        conv.log.set_verbose(arguments.verbose, arguments.quiet)

        filename, format_ = resolve_name(arguments, fname)
        if conv.save(filename, format_) != 0:
            exit_code += 1

    return exit_code


def resolve_name(arguments, fname):
    """
    Return right name and format for an output file.
    """
    if arguments.output:
        if len(arguments.filename) > 1:
            if not os.path.exists(arguments.output):
                os.mkdir(arguments.output)
            if not os.path.isdir(arguments.output):
                raise IOError("Path `%s' is not directory" % arguments.output)
            filename = os.path.join(arguments.output,
                                    get_modified_fname(fname, "prg"))
        else:
            filename = arguments.output
    else:
        filename = get_modified_fname(fname, "prg")

    format_ = arguments.format

    if hasattr(arguments, "executable") and arguments.executable:

        format_ = "prg"
        _, ext = os.path.splitext(filename)
        if ext != ".prg":
            filename = get_modified_fname(filename, "prg")

    if hasattr(arguments, "raw") and arguments.raw:

        format_ = "raw"
        filename, ext = os.path.splitext(filename)

    return filename, format_


def image2c64():
    """
    Parse options, run the conversion
    """
    class_map = {"art-studio-hires": HiresConverter,
                 "hires": HiresConverter,
                 "koala": MultiConverter,
                 "multi": MultiConverter}

    formatter = argparse.RawDescriptionHelpFormatter
    parser = argparse.ArgumentParser(description=__doc__,
                                     formatter_class=formatter)
    parser.add_argument("-g", "--border", help="set color number for border, "
                        "default: most frequent color", type=int,
                        choices=range(16))
    parser.add_argument("-b", "--background", help="set color number for "
                        "background", type=int, choices=range(16))
    parser.add_argument("-e", "--errors", help="perform the action in case of "
                        "color clashes: save errormap under the same name "
                        "with '_error' suffix, show it, open in grafx2, fix "
                        "it, or don't do anything (the message appear)",
                        default="none", choices=("show", "save", "grafx2",
                                                 "fix", "none"))
    parser.add_argument("-f", "--format", help="format of output file, this "
                        "option is mandatory", choices=class_map.keys(),
                        required=True)
    group = parser.add_mutually_exclusive_group()
    group.add_argument("-x", "--executable", help="produce C64 executable as"
                       " 'prg' file", action="store_true")
    group.add_argument("-r", "--raw", help="produce raw files with only the "
                       "data. Useful for include in assemblers",
                       action="store_true")
    parser.add_argument("-o", "--output", help="output filename, default: "
                        "same filename as original with appropriate extension"
                        ". If multiple files provided as the input, output "
                        "will be treated as the directory")
    parser.add_argument('filename', nargs="+")

    group = parser.add_mutually_exclusive_group()
    group.add_argument("-q", "--quiet", help='please, be quiet. Adding more '
                       '"q" will decrease verbosity', action="count",
                       default=0)
    group.add_argument("-v", "--verbose", help='be verbose. Adding more "v" '
                       'will increase verbosity', action="count", default=0)
    parser.add_argument("-V", "--version", action='version',
                        version="%(prog)s v" + ver)

    arguments = parser.parse_args()
    sys.exit(convert(arguments, class_map[arguments.format]))



#!/usr/bin/env python
"""
Tests for c64img.path module
This is part of the c64img project <https://bitbucket.org/gryf/image2c64>

Author: Roman 'gryf' Dobosz <gryf73@gmail.com>
Licence: BSD
"""
from unittest import TestCase, main

from c64img import path


class TestPath(TestCase):
    """
    Test helper functions
    """

    def test_get_modified_fname(self):
        """
        Test get_modified_fname function.
        """
        pic_path = "foo/hires.png"
        self.assertEqual(path.get_modified_fname(pic_path, "png", "_foo"),
                         "foo/hires_foo.png")
        self.assertEqual(path.get_modified_fname(pic_path, ".png", "_foo"),
                         "foo/hires_foo.png")
        self.assertEqual(path.get_modified_fname(pic_path, "png", "_foo."),
                         "foo/hires_foo.png")


if __name__ == "__main__":
    main()



#!/usr/bin/env python
"""
Tests for Image2c64 <https://bitbucket.org/gryf/image2c64>

Author: Roman 'gryf' Dobosz <gryf73@gmail.com>
Date: 2012-11-20
Version: 1.3
Licence: BSD
"""
from unittest import TestCase, main
import logging

from c64img import logger


class TestLogger(TestCase):
    """
    Test Logger class
    """
    def test___init(self):
        """
        Test logger object creation
        """
        log = logger.Logger("foo")
        self.assertTrue(isinstance(log, logger.Logger))
        # since logger.Logging is mocked here we will have DummyLogger
        # instance
        self.assertTrue(isinstance(log(), logging.Logger))  # DummyLogger))
        self.assertEqual(log().handlers[0].name, 'console')

    def test__setup_logger(self):
        """
        Test setup logger method
        """
        log = logger.Logger("foo")
        self.assertEqual(len(log().handlers), 1)
        log.setup_logger()
        self.assertEqual(len(log().handlers), 1)
        self.assertEqual(log().handlers[0].name, 'console')
        self.assertEqual(log().getEffectiveLevel(), logging.WARNING)

    def test_set_verbose(self):
        """
        Test for changing verbosity level
        """
        log = logger.Logger("foo")()

        # Default levels are set to 0, which means only warnings are visible
        log.set_verbose(0, 0)
        self.assertEqual(log.getEffectiveLevel(), logging.WARNING)

        # more verbose
        log.set_verbose(1, 0)
        self.assertEqual(log.getEffectiveLevel(), logging.INFO)

        # even more verbose
        log.set_verbose(2, 0)
        self.assertEqual(log.getEffectiveLevel(), logging.DEBUG)

        # further verbosity doesn't make sense
        log.set_verbose(5, 0)
        self.assertEqual(log.getEffectiveLevel(), logging.DEBUG)

        # less verbose
        log.set_verbose(0, 1)
        self.assertEqual(log.getEffectiveLevel(), logging.ERROR)

        # even less verbose
        log.set_verbose(0, 2)
        self.assertEqual(log.getEffectiveLevel(), logging.CRITICAL)

        # smaller verbosity doesn't make sense
        log.set_verbose(0, 5)
        self.assertEqual(log.getEffectiveLevel(), logging.CRITICAL)

        # even though both values cannot be set at the same time (argparse
        # module is taking care about that), verbosity takes precedence over
        # quietness.
        log.set_verbose(5, 5)
        self.assertEqual(log.getEffectiveLevel(), logging.DEBUG)


if __name__ == "__main__":
    main()



"""
File and path manipulation utility function(s)
"""
import os


def get_modified_fname(fname, ext, suffix='.'):
    """
    Change the name of provided filename to different. Suffix should contain
    dot, since it is last part of the filename and dot should separate it
    from extension. If not, dot will be added automatically.
    """
    path, _ = os.path.splitext(fname)
    if not (suffix.endswith(".") or ext.startswith(".")):
        ext = "." + ext
    return "".join([path, suffix, ext])



#!/usr/bin/env python
"""
Tests for c64img.multi module
This is part of the c64img project <https://bitbucket.org/gryf/image2c64>

Author: Roman 'gryf' Dobosz <gryf73@gmail.com>
Licence: BSD
"""
import os
from unittest import TestCase, main
from tempfile import mkstemp

from PIL import Image

from c64img import multi


Image.Image.show = lambda x: None


def get_image_path(image_name):
    return os.path.join(os.path.dirname(__file__), "test_images", image_name)


CLASH_H = get_image_path("clash.hires.png")
CLASH_M = get_image_path("clash.multi.png")
CLASH_M2 = get_image_path("clash.multi.160.png")
COLORS_256 = get_image_path("colors.256.png")
PAL_PEPTO = get_image_path("colors.pepto.png")
PAL_TIMANTHES = get_image_path("colors.timanthes.png")
PAL_UNKNOWN = get_image_path("colors.unknown.png")
PAL_VICE = get_image_path("colors.vice.png")
CROP_BOTH = get_image_path("crop.251x187.png")
CROP_H = get_image_path("crop.319x200.png")
CROP_V = get_image_path("crop.320x199.png")
COLORS_1 = get_image_path("hires_1c.png")
COLORS_2 = get_image_path("hires_2c.png")
COLORS_256_U16 = get_image_path("hires_256_16p.png")  # 256 defined, uniq 16
HIRES = get_image_path("hires.png")
MULTI = get_image_path("multi.160x200.png")
MULTI_320 = get_image_path("multi.320x200.png")


class LogMock:
    """Mock logger class"""

    def warning(*args, **kwargs):
        return

    def debug(*args, **kwargs):
        return

    def critical(*args, **kwargs):
        return

    def getEffectiveLevel(self):
        return 30  # warn


class Interceptor:
    """
    Interceptor class for function call detection
    """
    def __init__(self):
        """
        Init.
        """
        self.call = 0
        self.repeat = False

    def __call__(self, dummy1, dummy2=False):
        """
        Call attribute is increased every time, instance is called
        """
        self.call += 1
        return self.repeat


class TestMultiChar(TestCase):
    """
    Test multicolor Char class
    """
    def setUp(self):
        """set up"""
        self.log = LogMock()

    def test___init(self):
        """
        Test Char initialization
        """
        char = multi.MultiChar(self.log, {0: (0, 0)})
        self.assertEqual(char.max_colors, 4)
        self.assertEqual(char.pixel_state, {(0, 1): False,
                                            (1, 0): False,
                                            (1, 1): False})

    def test_get_binary_data(self):
        """
        Test get_binary_data method
        """
        char = multi.MultiChar(self.log, {0: (0, 0)})
        char.pixels = {(0, 0): 0, (0, 1): 1, (0, 2): 2, (0, 3): 3,
                       (1, 0): 0, (1, 1): 1, (1, 2): 2, (1, 3): 3}
        char._analyze_color_map()
        result = char.get_binary_data()
        self.assertEqual(result['bitmap'], [57, 0b111001])  # 57 for all
        self.assertEqual(result['color-ram'], 50)
        self.assertEqual(result['screen-ram'], 1)

        # last pixel with the clash - should fall back to background color
        char.pixels = {(0, 0): 0, (0, 1): 1, (0, 2): 2, (0, 3): 3,
                       (1, 0): 0, (1, 1): 1, (1, 2): 2, (1, 3): 3,
                       (2, 0): 0, (2, 1): 1, (2, 2): 2, (2, 3): 3,
                       (3, 0): 0, (3, 1): 1, (3, 2): 2, (3, 3): 3,
                       (4, 0): 0, (4, 1): 1, (4, 2): 2, (4, 3): 3,
                       (5, 0): 0, (5, 1): 1, (5, 2): 2, (5, 3): 3,
                       (6, 0): 0, (6, 1): 1, (6, 2): 2, (6, 3): 3,
                       (7, 0): 0, (7, 1): 1, (7, 2): 2, (7, 3): 9}
        result = char.get_binary_data()
        self.assertEqual(result['bitmap'], [57, 57, 57, 57, 57, 57, 57,
                                            0b111000])
        self.assertEqual(result['color-ram'], 50)
        self.assertEqual(result['screen-ram'], 1)

    def test_analyze_color_map(self):
        """
        Test _analyze_color_map method
        """
        char = multi.MultiChar(self.log, {0: (0, 0)})
        # simulate clash
        char.max_colors = 1
        char.pixels[(0, 0)] = 1
        char.pixels[(0, 1)] = 2
        char._analyze_color_map()
        self.assertEqual(char.colors[0], (0, 0))

        # simulate previous image clash
        char.clash = False
        char.max_colors = 2
        char.prev = multi.MultiChar(self.log, {0: (0, 0)})
        char.prev.clash = True
        char._analyze_color_map()
        self.assertEqual(char.colors[0], (0, 0))
        self.assertEqual(len(char.colors), 1)

    def test__compare_colors_with_prev_char(self):
        """
        Test _compare_colors_with_prev_char method. This method is responsible
        for creating new mapping for multicolor bit pairs in conjunction with
        corresponding colors.
        """
        char = multi.MultiChar(self.log, {0: (0, 0)})
        colors = {0: 16,
                  1: 4,
                  2: 5,
                  3: 7}

        # The default case. No previous picture stored, no colors to compare
        # with. None colors were stored.
        self.assertTrue(char._compare_colors_with_prev_char(colors))
        self.assertEqual(char.pixel_state, {(1, 0): False,
                                            (0, 1): False,
                                            (1, 1): False})
        self.assertEqual(char.colors, {})

        # So it needs to rerun, colors are recognized and remembered.
        self.assertFalse(char._compare_colors_with_prev_char(colors, True))
        self.assertEqual(char.pixel_state, {(1, 0): True,
                                            (0, 1): True,
                                            (1, 1): True})
        self.assertEqual(char.colors, {1: (0, 1), 2: (1, 0), 3: (1, 1)})

        # 1. Ideal case. Colors for previous and current character in char
        # boundary are the same. No need to rerun checks.
        char = multi.MultiChar(self.log, {0: (0, 0)})
        prev = multi.MultiChar(self.log, {0: (0, 0)})
        prev.colors = {1: (0, 1), 2: (1, 0), 3: (1, 1)}
        char.prev = prev
        self.assertFalse(char._compare_colors_with_prev_char(colors))
        self.assertEqual(char.pixel_state, {(1, 0): True,
                                            (0, 1): True,
                                            (1, 1): True})
        self.assertEqual(char.colors, {1: (0, 1), 2: (1, 0), 3: (1, 1)})

        # 2. Mixed colors/pixel pairs. Color indices are matching fine. Colors
        # and pairs from previous character should be propagated into current
        # char.
        char = multi.MultiChar(self.log, {0: (0, 0)})
        prev = multi.MultiChar(self.log, {0: (0, 0)})
        prev.colors = {3: (0, 1), 1: (1, 0), 2: (1, 1)}
        char.prev = prev
        self.assertFalse(char._compare_colors_with_prev_char(colors))
        self.assertEqual(char.pixel_state, {(1, 0): True,
                                            (0, 1): True,
                                            (1, 1): True})
        self.assertEqual(char.colors, {3: (0, 1), 1: (1, 0), 2: (1, 1)})

        # 3. Mixed colors/pixel pairs. One color index differ. Colors and
        # pairs from previous character should be propagated into current
        # char, the mismatch color should be replaced by current one.
        char = multi.MultiChar(self.log, {0: (0, 0)})
        prev = multi.MultiChar(self.log, {0: (0, 0)})
        prev.colors = {3: (0, 1), 4: (1, 0), 2: (1, 1)}
        char.prev = prev
        self.assertTrue(char._compare_colors_with_prev_char(colors))
        self.assertEqual(char.pixel_state, {(1, 0): False,
                                            (0, 1): True,
                                            (1, 1): True})
        self.assertEqual(char.colors, {3: (0, 1), 2: (1, 1)})
        self.assertFalse(char._compare_colors_with_prev_char(colors, True))
        self.assertEqual(char.pixel_state, {(1, 0): True,
                                            (0, 1): True,
                                            (1, 1): True})
        self.assertEqual(char.colors, {3: (0, 1), 2: (1, 1), 1: (1, 0)})

        # 4. Mixed colors/pixel pairs. One color index match. Colors and
        # pairs from previous character should be propagated into current
        # char, the mismatch colors should be replaced by current ones.
        char = multi.MultiChar(self.log, {0: (0, 0)})
        prev = multi.MultiChar(self.log, {6: (0, 0)})
        prev.colors = {4: (0, 1), 3: (1, 0), 5: (1, 1)}
        char.prev = prev
        self.assertTrue(char._compare_colors_with_prev_char(colors))
        self.assertEqual(char.pixel_state, {(1, 0): True,
                                            (0, 1): False,
                                            (1, 1): False})
        self.assertEqual(char.colors, {3: (1, 0)})
        self.assertFalse(char._compare_colors_with_prev_char(colors, True))
        self.assertEqual(char.pixel_state, {(1, 0): True,
                                            (0, 1): True,
                                            (1, 1): True})
        self.assertEqual(char.colors, {3: (1, 0), 1: (0, 1), 2: (1, 1)})

        # 5. Worst case scenario. None of the colors from previous char
        # matches. Get the current colors.
        char = multi.MultiChar(self.log, {0: (0, 0)})
        prev = multi.MultiChar(self.log, {6: (0, 0)})
        prev.colors = {4: (0, 1), 5: (1, 0), 6: (1, 1)}
        char.prev = prev
        self.assertTrue(char._compare_colors_with_prev_char(colors))
        self.assertEqual(char.pixel_state, {(1, 0): False,
                                            (0, 1): False,
                                            (1, 1): False})
        self.assertEqual(char.colors, {})
        self.assertFalse(char._compare_colors_with_prev_char(colors, True))
        self.assertEqual(char.pixel_state, {(1, 0): True,
                                            (0, 1): True,
                                            (1, 1): True})
        self.assertEqual(char.colors, {1: (0, 1), 2: (1, 0), 3: (1, 1)})

    def test__fix_color_clash(self):
        """Test for repait color clash"""
        bg = 0  # black
        col1 = 1  # white
        col2 = 2  # red
        col3 = 3  # magenta
        clash = 4  # purple

        char = multi.MultiChar(self.log, {bg: (0, 0)}, prev=None,
                               fix_clash=True)

        char.pixels = {(0, 0): bg, (0, 1): col1, (0, 2): col2, (0, 3): col3,
                       (1, 0): bg, (1, 1): col1, (1, 2): col2, (1, 3): col3,
                       (2, 0): bg, (2, 1): col1, (2, 2): col2, (2, 3): col3,
                       (3, 0): bg, (3, 1): col1, (3, 2): col2, (3, 3): col3,
                       (4, 0): bg, (4, 1): col1, (4, 2): col2, (4, 3): col3,
                       (5, 0): bg, (5, 1): col1, (5, 2): col2, (5, 3): col3,
                       (6, 0): bg, (6, 1): col1, (6, 2): col2, (6, 3): col3,
                       (7, 0): bg, (7, 1): col1, (7, 2): col2, (7, 3): clash}

        char._analyze_color_map()
        self.assertEqual(char.pixels[(7, 3)], col2)

        c1 = 3  # magenta
        c2 = 4  # purple
        c3 = 5  # green
        c4 = 6  # d.blue

        char = multi.MultiChar(self.log, {bg: (0, 0)}, prev=None,
                               fix_clash=True)
        char.pixels = {(0, 0): c1, (0, 1): c2, (0, 2): c3, (0, 3): c4,
                       (1, 0): c1, (1, 1): c2, (1, 2): c3, (1, 3): c4,
                       (2, 0): c1, (2, 1): c2, (2, 2): c3, (2, 3): c4,
                       (3, 0): c1, (3, 1): c2, (3, 2): c3, (3, 3): c4,
                       (4, 0): c1, (4, 1): c2, (4, 2): c3, (4, 3): c4,
                       (5, 0): c1, (5, 1): c2, (5, 2): c3, (5, 3): c4,
                       (6, 0): c1, (6, 1): c2, (6, 2): c3, (6, 3): c4,
                       (7, 0): c1, (7, 1): c2, (7, 2): c3, (7, 3): c4}

        char._analyze_color_map()
        for idx in range(8):
            self.assertEqual(char.pixels[(idx, 0)], c1)
            self.assertEqual(char.pixels[(idx, 1)], c2)
            self.assertEqual(char.pixels[(idx, 2)], c3)
            self.assertEqual(char.pixels[(idx, 3)], bg)

        c1 = 1  # white - clash
        c2 = 2  # red
        c3 = 5  # green
        c4 = 3  # magenta

        char = multi.MultiChar(self.log, {bg: (0, 0)}, prev=None,
                               fix_clash=True)
        char.pixels = {(0, 0): c1, (0, 1): c2, (0, 2): c3, (0, 3): c4,
                       (1, 0): c1, (1, 1): c2, (1, 2): c3, (1, 3): c4,
                       (2, 0): c1, (2, 1): c2, (2, 2): c3, (2, 3): c4,
                       (3, 0): bg, (3, 1): c2, (3, 2): c3, (3, 3): c4,
                       (4, 0): bg, (4, 1): c2, (4, 2): c3, (4, 3): c4,
                       (5, 0): bg, (5, 1): c2, (5, 2): c3, (5, 3): c4,
                       (6, 0): bg, (6, 1): c2, (6, 2): c3, (6, 3): c4,
                       (7, 0): bg, (7, 1): c2, (7, 2): c3, (7, 3): c4}

        char._analyze_color_map()
        for idx in range(8):
            self.assertEqual(char.pixels[(idx, 0)], bg)
            self.assertEqual(char.pixels[(idx, 1)], c2)
            self.assertEqual(char.pixels[(idx, 2)], c3)
            self.assertEqual(char.pixels[(idx, 3)], c4)


class TestMulticolor(TestCase):
    """
    Tests for multicolor conversion
    """
    def test__check_dimensions(self):
        """
        Test of dimensions of the input images
        """
        for fname in (CROP_V, CROP_H, CROP_BOTH):
            obj = multi.MultiConverter(fname)
            obj.log.error = lambda *x: None  # suppress log
            obj._load()
            self.assertFalse(obj._check_dimensions())

        for fname in (CLASH_H, CLASH_M, COLORS_256, COLORS_1, COLORS_2,
                      COLORS_256_U16, HIRES, MULTI_320, CLASH_M2, MULTI,
                      PAL_PEPTO, PAL_TIMANTHES, PAL_UNKNOWN, PAL_VICE):
            obj = multi.MultiConverter(fname)
            obj._load()
            self.assertTrue(obj._check_dimensions(), fname)

    def test__load(self):
        """
        Test custom load function
        """
        obj = multi.MultiConverter(MULTI)
        self.assertEqual(obj._load(), True)
        self.assertEqual(obj._src_image.size[0], 160)

        obj = multi.MultiConverter(MULTI_320)
        self.assertEqual(obj._load(), True)
        self.assertEqual(obj._src_image.size[0], 160)

        obj = multi.MultiConverter("nofile")
        obj.log = LogMock()
        self.assertEqual(obj._load(), False)
        self.assertEqual(obj._src_image, None)

    def test_save(self):
        """
        Test for save methods
        """
        fdesc, fname = mkstemp(suffix=".prg")
        os.close(fdesc)
        os.unlink(fname)

        obj = multi.MultiConverter(MULTI_320)
        self.assertEqual(obj.save(fname, "prg"), 0)

        self.assertTrue(os.path.exists(fname))
        self.assertEqual(os.stat(fname).st_size, 14145)
        os.unlink(fname)

        fdesc, fname = mkstemp(suffix=".prg")
        os.close(fdesc)
        os.unlink(fname)

        obj = multi.MultiConverter(MULTI)
        self.assertEqual(obj.save(fname, "koala"), 0)

        self.assertTrue(os.path.exists(fname))
        self.assertEqual(os.stat(fname).st_size, 10003)
        os.unlink(fname)

        obj = multi.MultiConverter(CLASH_M)
        self.assertEqual(obj.save(fname, "koala"), 1)

        obj = multi.MultiConverter(CLASH_M2)
        self.assertEqual(obj.save(fname, "koala"), 1)

        obj = multi.MultiConverter(MULTI)
        self.assertEqual(obj.save(fname, "raw"), 0)

        self.assertTrue(os.path.exists(fname + "_bg.raw"))
        self.assertTrue(os.path.exists(fname + "_bitmap.raw"))
        self.assertTrue(os.path.exists(fname + "_color-ram.raw"))
        self.assertTrue(os.path.exists(fname + "_screen.raw"))

        self.assertEqual(os.stat(fname + "_bg.raw").st_size, 1)
        self.assertEqual(os.stat(fname + "_bitmap.raw").st_size, 8000)
        self.assertEqual(os.stat(fname + "_color-ram.raw").st_size, 1000)
        self.assertEqual(os.stat(fname + "_screen.raw").st_size, 1000)
        os.unlink(fname + "_bg.raw")
        os.unlink(fname + "_bitmap.raw")
        os.unlink(fname + "_color-ram.raw")
        os.unlink(fname + "_screen.raw")

        obj = multi.MultiConverter(CLASH_M)
        self.assertEqual(obj.save(fname, "raw"), 1)

        obj = multi.MultiConverter(CLASH_M2)
        self.assertEqual(obj.save(fname, "raw"), 1)


if __name__ == "__main__":
    main()



#!/usr/bin/env python
"""
Tests for c64img module <https://bitbucket.org/gryf/base>

Author: Roman 'gryf' Dobosz <gryf73@gmail.com>
Date: 2012-11-20
Version: 1.3
Licence: BSD
"""
import os
from unittest import TestCase, main
import logging

from PIL import Image

from c64img import base


Image.Image.show = lambda x: None


def get_image_path(image_name):
    return os.path.join(os.path.dirname(__file__), "test_images", image_name)


CLASH_H = get_image_path("clash.hires.png")
CLASH_M = get_image_path("clash.multi.png")
CLASH_M2 = get_image_path("clash.multi.160.png")
COLORS_256 = get_image_path("colors.256.png")
PAL_PEPTO = get_image_path("colors.pepto.png")
PAL_TIMANTHES = get_image_path("colors.timanthes.png")
PAL_UNKNOWN = get_image_path("colors.unknown.png")
PAL_VICE = get_image_path("colors.vice.png")
CROP_BOTH = get_image_path("crop.251x187.png")
CROP_H = get_image_path("crop.319x200.png")
CROP_V = get_image_path("crop.320x199.png")
COLORS_1 = get_image_path("hires_1c.png")
COLORS_2 = get_image_path("hires_2c.png")
COLORS_256_U16 = get_image_path("hires_256_16p.png")  # 256 defined, uniq 16
HIRES = get_image_path("hires.png")
MULTI = get_image_path("multi.160x200.png")
MULTI_320 = get_image_path("multi.320x200.png")


class LogMock:
    """Mock logger class"""
    def warning(*args, **kwargs):
        return

    def debug(*args, **kwargs):
        return

    def critical(*args, **kwargs):
        return


class Interceptor:
    """
    Interceptor class for function call detection
    """
    def __init__(self):
        """
        Init.
        """
        self.call = 0
        self.repeat = False

    def __call__(self, dummy1, dummy2=False):
        """
        Call attribute is increased every time, instance is called
        """
        self.call += 1
        return self.repeat


class TestChar(TestCase):
    """
    Test Char class
    """
    def setUp(self):
        """set up"""
        self.log = LogMock()

    def test___init(self):
        """
        Test Char initialization
        """
        char = base.Char(self.log)
        self.assertEqual(char.clash, False)
        self.assertEqual(char.colors, {})
        self.assertEqual(char.max_colors, 2)
        self.assertEqual(char.pixels, {})
        self.assertEqual(char.prev, None)

        char = base.Char(self.log, "foo")
        self.assertEqual(char.prev, "foo")

    def test__analyze_color_map(self):
        """
        Test _analyze_color_map method
        """
        char = base.Char(self.log, 0)
        # simulate clash
        char.max_colors = 1
        char.pixels[(0, 0)] = 1
        char.pixels[(0, 1)] = 2
        char._analyze_color_map()
        self.assertEqual(char.colors, {})

        # simulate previous image clash
        char.clash = False
        char.max_colors = 2
        char.prev = base.Char(self.log, 0)
        char.prev.clash = True
        char._analyze_color_map()
        self.assertEqual(char.colors, {})
        self.assertEqual(len(char.colors), 0)

    def test_get_binary_data(self):
        """
        Test get_binary_data method
        """
        char = base.Char(self.log, 0)
        self.assertRaises(NotImplementedError, char.get_binary_data)

    def test__check_clash(self):
        """
        Test _check_clash method
        """
        char = base.Char(self.log, 0)
        self.assertEqual(char._check_clash(), False)

    def test__compare_colors(self):
        """
        Test _compare_colors method
        """
        char = base.Char(self.log)
        self.assertRaises(TypeError, char._compare_colors)
        self.assertRaises(TypeError, char._compare_colors, None)
        char._compare_colors([])


class TestFullScreenImage(TestCase):
    """
    Test FullScreenImage class
    """
    def setUp(self):
        """setup"""
        self._system = os.system
        self.os_system_params = ""

        def _system(param):
            self.os_system_params = param

        os.system = _system

    def tearDown(self):
        """Restore stuff"""
        os.system = self._system

    def test___init(self):
        """
        Test initialization
        """
        self.assertRaises(TypeError, base.FullScreenImage)
        obj = base.FullScreenImage(MULTI_320)
        self.assertEqual(obj._fname, MULTI_320)
        self.assertEqual(obj._errors_action, "none")
        self.assertEqual(obj._src_image, None)
        self.assertTrue(isinstance(obj.log, logging.Logger))
        self.assertEqual(obj.data, {})

    def test__check_dimensions(self):
        """
        Test of dimensions of the input images
        """
        for fname in (CROP_V, CROP_H, CROP_BOTH, CLASH_M2, MULTI, PAL_PEPTO,
                      PAL_TIMANTHES, PAL_UNKNOWN, PAL_VICE):
            obj = base.FullScreenImage(fname)
            obj.log.error = lambda *x: None  # suppress log
            obj._load()
            self.assertFalse(obj._check_dimensions())

        for fname in (CLASH_H, CLASH_M, COLORS_256, COLORS_1, COLORS_2,
                      COLORS_256_U16, HIRES, MULTI_320):
            obj = base.FullScreenImage(fname)
            obj._load()
            self.assertTrue(obj._check_dimensions())

    def test__load(self):
        """
        Test of loading images
        """
        self.assertTrue(base.FullScreenImage(HIRES)._load())
        obj = base.FullScreenImage("nofile")
        self.assertFalse(obj._load())

        obj = base.FullScreenImage("/none/existing")
        self.assertFalse(obj._load())
        obj.log.setLevel(logging.DEBUG)
        self.assertRaises(IOError, obj._load)
        obj.log.setLevel(logging.WARNING)

    def test__colors_check(self):
        """
        Test _colors_check method
        """
        obj = base.FullScreenImage(COLORS_256)
        obj._load()
        histogram = obj._src_image.histogram()
        # even if input image have 256 colors, internally it will be converted
        # to 16 colors.
        self.assertEqual(obj._colors_check(histogram), 16)

        obj = base.FullScreenImage(COLORS_1)
        obj.log.warn = lambda x, y: None  # suppress log
        obj._load()
        histogram = obj._src_image.histogram()
        self.assertEqual(obj._colors_check(histogram), 1)

        obj = base.FullScreenImage(COLORS_2)
        obj._load()
        histogram = obj._src_image.histogram()
        self.assertEqual(obj._colors_check(histogram), 2)

    def test_convert(self):
        """
        Test convert process
        """
        obj = base.FullScreenImage("nofile")
        self.assertFalse(obj.convert())

        obj = base.FullScreenImage(COLORS_256)
        self.assertRaises(NotImplementedError, obj.convert)

        obj = base.FullScreenImage(CROP_BOTH)
        self.assertFalse(obj.convert())

    def test__get_best_palette_map(self):
        """
        Test for method _find_best_palette_map
        """
        obj = base.FullScreenImage(COLORS_1)
        obj._load()
        obj._find_best_palette_map()
        self.assertEqual(obj._palette_map, {(0, 0, 0): 0})

        obj = base.FullScreenImage(COLORS_2)
        obj._load()
        obj._find_best_palette_map()
        histogram = obj._src_image.histogram()
        obj._find_most_freq_color(histogram)

        self.assertEqual(obj.data['most_freq_color'], 11)

    def test__find_most_freq_color(self):
        """
        Test for method _find_most_freq_color
        """
        obj = base.FullScreenImage(COLORS_2)
        self.assertEqual(obj.data.get('most_freq_color'), None)

        obj._load()
        obj._find_best_palette_map()
        histogram = obj._src_image.histogram()
        obj._find_most_freq_color(histogram)

        self.assertEqual(obj.data['most_freq_color'], 11)

        obj = base.FullScreenImage(COLORS_1)
        self.assertEqual(obj.data.get('most_freq_color'), None)

        obj._load()
        obj._find_best_palette_map()
        histogram = obj._src_image.histogram()
        obj._find_most_freq_color(histogram)

        self.assertEqual(obj.data['most_freq_color'], 0)

    def test__get_palette(self):
        """
        Test _get_palette method
        """
        obj = base.FullScreenImage(PAL_PEPTO)
        obj._load()

        ref_palette = set(((0, 0, 0), (53, 40, 121), (67, 57, 0),
                           (68, 68, 68), (88, 141, 67), (104, 55, 43),
                           (108, 94, 181), (108, 108, 108), (111, 61, 134),
                           (111, 79, 37), (112, 164, 178), (149, 149, 149),
                           (154, 103, 89), (154, 210, 132), (184, 199, 111),
                           (255, 255, 255)))

        # got 16 colors pepto palette ordered by lightness
        self.assertEqual(set(obj._get_palette()), ref_palette)

    def test__discover_best_palette(self):
        """
        Test _discover_best_palette method
        """
        obj = base.FullScreenImage(PAL_PEPTO)
        obj._load()
        obj._find_best_palette_map()
        self.assertEqual(obj._palette_map,
                         {(184, 199, 111): 7, (154, 210, 132): 13,
                          (112, 164, 178): 3, (88, 141, 67): 5,
                          (0, 0, 0): 0, (108, 94, 181): 14,
                          (68, 68, 68): 11, (154, 103, 89): 10,
                          (67, 57, 0): 9, (53, 40, 121): 6,
                          (111, 79, 37): 8, (108, 108, 108): 12,
                          (111, 61, 134): 4, (255, 255, 255): 1,
                          (104, 55, 43): 2, (149, 149, 149): 15})

        obj = base.FullScreenImage(PAL_VICE)
        obj._load()
        obj._find_best_palette_map()
        self.assertEqual(obj._palette_map,
                         {(182, 89, 0): 9, (0, 142, 0): 5,
                          (147, 81, 182): 4, (56, 255, 52): 5,
                          (154, 154, 154): 15, (0, 81, 158): 6,
                          (134, 134, 134): 12, (0, 182, 182): 14,
                          (255, 109, 109): 10, (0, 0, 0): 0,
                          (85, 85, 85): 11, (109, 121, 255): 14,
                          (213, 223, 124): 7, (255, 255, 255): 1,
                          (109, 52, 0): 9, (207, 0, 0): 2})

        obj = base.FullScreenImage(PAL_TIMANTHES)
        obj._load()
        obj._find_best_palette_map()
        self.assertEqual(obj._palette_map,
                         {(103, 93, 182): 14,
                          (143, 194, 113): 13,
                          (86, 141, 53): 5, (174, 183, 94): 7,
                          (101, 159, 166): 3, (143, 143, 143): 15,
                          (156, 100, 90): 10, (115, 58, 145): 4,
                          (0, 0, 0): 0, (71, 71, 71): 11,
                          (75, 60, 0): 9, (114, 53, 44): 2,
                          (107, 107, 107): 12, (119, 79, 30): 8,
                          (213, 213, 213): 1, (46, 35, 125): 6})

        obj = base.FullScreenImage(PAL_UNKNOWN)
        obj._load()
        obj._find_best_palette_map()
        self.assertEqual(obj._palette_map,
                         {(91, 50, 8): 9, (130, 130, 130): 12,
                          (230, 122, 122): 10, (0, 181, 181): 14,
                          (211, 211, 112): 7, (23, 63, 194): 6,
                          (49, 150, 49): 5, (79, 123, 255): 14,
                          (198, 0, 0): 2, (193, 90, 0): 9,
                          (160, 160, 160): 15, (178, 66, 178): 4,
                          (255, 255, 255): 1, (42, 232, 112): 5,
                          (85, 85, 85): 11, (0, 0, 0): 0})

    def test__fill_memory(self):
        """
        Test _fill_memory method
        """
        obj = base.FullScreenImage(HIRES)
        obj._load()
        self.assertRaises(NotImplementedError, obj._fill_memory)

    def test_save(self):
        """
        Test save method
        """
        obj = base.FullScreenImage(HIRES)
        obj._load()
        self.assertRaises(NotImplementedError, obj.save, "fn")

    def test_attributes(self):
        """
        Test settting background/border colors
        """
        obj = base.FullScreenImage(HIRES)
        self.assertEqual(obj.data, {})

        obj.set_border_color(12)
        self.assertEqual(obj.data, {'border': 12})

        obj.set_bg_color(11)
        self.assertEqual(obj.data, {'background': 11, 'border': 12})

        obj.set_border_color(1)
        self.assertEqual(obj.data, {'background': 11, 'border': 1})

        obj.set_bg_color(0)
        self.assertEqual(obj.data, {'background': 0, 'border': 1})

    def test__get_border(self):
        """
        Test for _get_border method
        """
        obj = base.FullScreenImage(HIRES)
        self.assertEqual(obj._get_border(), 0)

        obj._load()
        obj._find_best_palette_map()
        histogram = obj._src_image.histogram()
        obj._find_most_freq_color(histogram)
        self.assertEqual(obj._get_border(), 0)

        obj = base.FullScreenImage(COLORS_256)
        obj.set_border_color(12)
        self.assertEqual(obj._get_border(), 12)

        obj = base.FullScreenImage(COLORS_256)
        self.assertEqual(obj._get_border(), 0)

        obj._load()
        obj._find_best_palette_map()
        histogram = obj._src_image.histogram()
        obj._find_most_freq_color(histogram)
        self.assertEqual(obj._get_border(), 5)
        obj.set_border_color(7)
        self.assertEqual(obj._get_border(), 7)

    def test_set_bg_color(self):
        """
        Test set_bg_color method
        """
        obj = base.FullScreenImage(HIRES)
        self.assertEqual(obj.data.get('background'), None)
        obj.set_bg_color(15)
        self.assertEqual(obj.data.get('background'), 15)

    def test_set_border_color(self):
        """
        Test set_border_color method
        """
        obj = base.FullScreenImage(HIRES)
        self.assertEqual(obj.data.get('border'), None)
        obj.set_border_color(14)
        self.assertEqual(obj.data.get('border'), 14)

    def test__error_image_action(self):
        """
        Test for _error_image_action method
        """
        error_img = base.get_modified_fname(MULTI, 'png', '_error.')

        if os.path.exists(error_img):
            os.unlink(error_img)

        obj = base.FullScreenImage(MULTI)
        obj._load()
        result = obj._error_image_action([(0, 0)], True)
        self.assertEqual(result, None)

        obj._errors_action = "save"
        result = obj._error_image_action([(0, 0)], True)
        self.assertTrue(os.path.exists(error_img))
        os.unlink(error_img)

        # does nothing. maybe some refactoring needed for _error_image_action
        # method?
        obj._errors_action = "show"
        obj._error_image_action([(0, 0)], True)

        obj = base.FullScreenImage(MULTI)
        obj._load()
        obj._errors_action = "show"
        obj._error_image_action([(0, 0)], True)

        error_img = base.get_modified_fname(MULTI_320, 'png', '_error.')

        if os.path.exists(error_img):
            os.unlink(error_img)

        obj = base.FullScreenImage(MULTI_320)
        obj._load()
        result = obj._error_image_action([(0, 0)])
        self.assertEqual(result, None)

        obj._errors_action = "save"
        result = obj._error_image_action([(0, 0)])
        self.assertTrue(os.path.exists(error_img))
        os.unlink(error_img)

        # does nothing. maybe some refactoring needed for _error_image_action
        # method?
        obj._errors_action = "show"
        obj._error_image_action([(0, 0)])

        obj = base.FullScreenImage(MULTI_320)
        obj._load()
        obj._errors_action = "show"
        obj._error_image_action([(0, 0)])

        error_img = base.get_modified_fname(HIRES, 'png', '_error.')

        if os.path.exists(error_img):
            os.unlink(error_img)

        obj = base.FullScreenImage(HIRES)
        obj._load()
        result = obj._error_image_action([(0, 0)])
        self.assertEqual(result, None)

        obj._errors_action = "save"
        result = obj._error_image_action([(0, 0)])
        self.assertTrue(os.path.exists(error_img))
        os.unlink(error_img)

        # does nothing. maybe some refactoring needed for _error_image_action
        # method?
        obj._errors_action = "show"
        obj._error_image_action([(0, 0)])

        obj = base.FullScreenImage(MULTI)
        obj._load()
        obj._errors_action = "show"
        obj._error_image_action([(0, 0)])

        # Test the grafx2 option
        error_img = base.get_modified_fname(MULTI_320, 'png', '_error.')
        obj = base.FullScreenImage(MULTI_320)
        obj._load()
        obj._errors_action = "grafx2"
        obj._error_image_action([(0, 0)])
        self.assertTrue("grafx2" in self.os_system_params)
        os.unlink(error_img)


class TestMisc(TestCase):
    """
    Test helper functions
    """

    def test_best_color_match(self):
        """
        Test _best_color_match helper function.
        """
        # perfect match
        idx, delta = base._best_color_match((0, 0, 0),
                                            base.PALETTES['Pepto'])
        self.assertEqual(idx, 0)
        self.assertEqual(delta, 0)

        # almost there
        idx, delta = base._best_color_match((2, 0, 0),
                                            base.PALETTES['Pepto'])
        self.assertEqual(idx, 0)
        self.assertEqual(delta, 4)

        idx, delta = base._best_color_match((254, 254, 254),
                                            base.PALETTES['Pepto'])
        self.assertEqual(idx, 1)
        self.assertEqual(delta, 3)

        # quite distorted: after all we got gray - which have values
        # 68, 68, 68
        idx, delta = base._best_color_match((86, 86, 86),
                                            base.PALETTES['Pepto'])
        self.assertEqual(idx, 11)
        self.assertEqual(delta, 972)

        # one value lower and now we got dark gray - which have values
        # 68, 68, 68
        idx, delta = base._best_color_match((85, 85, 85),
                                            base.PALETTES['Pepto'])
        self.assertEqual(idx, 11)
        self.assertEqual(delta, 867)


if __name__ == "__main__":
    main()



"""
Fullscreen multicolor image abstraction.

Assumptions:
    - Input image 160x200 or 320x200 pixels. In case od 320 px width images,
      it will be resized to 160x200 internally.
    - Maximum 3 colors per char may be used, with one constant color
      (background) available across all of the chars
"""
from c64img import base


class MultiChar(base.Char):
    """Char implementation for multicolor mode."""

    def __init__(self, log, pixels_pairs, prev=None, fix_clash=False):
        """
        Init. prev is the Char object which represents the same character
        in previous picture
        """
        super(MultiChar, self).__init__(log, prev, fix_clash)
        self.background = [x
                           for x, y in pixels_pairs.items()
                           if y == (0, 0)][0]
        self.max_colors = 4
        self.global_pixel_pairs = pixels_pairs  # 1: (0, 1), 2: (1, 1)...
        self.pixel_state = {(0, 1): False,
                            (1, 0): False,
                            (1, 1): False}

    def _analyze_color_map(self):
        """
        Check for the optimal color placement in char. This method may be run
        only on not clashed chars. Background color should be always
        available.
        """
        self.colors[self.background] = (0, 0)
        super(MultiChar, self)._analyze_color_map()

    def _compare_colors(self, colors):
        """Overwritten compare colors routine"""

        for color in colors:
            pair = self.global_pixel_pairs.get(color)
            if pair is not None:
                self.pixel_state[pair] = True
                self.colors[color] = pair

        for color in colors:
            if color == self.background:
                continue
            if color in self.global_pixel_pairs:
                continue

            if not self.pixel_state[(1, 1)]:
                self.pixel_state[(1, 1)] = True
                self.colors[color] = (1, 1)
                continue

            if not self.pixel_state[(1, 0)]:
                if self.prev and self.prev.colors.get(color, None) is not None:
                    # XXX: how about case, where we have previous data vs
                    # global colors prediction. Have to checkit.
                    pass
                self.log.debug("Anomaly reserving state 1,0, %s", self.prev)
                self.pixel_state[(1, 0)] = True
                self.colors[color] = (1, 0)
                continue

            if not self.pixel_state[(0, 1)]:
                self.log.debug("Anomaly reserving state 0,1")
                self.pixel_state[(0, 1)] = True
                self.colors[color] = (0, 1)
                continue

        for color in colors:
            if color in self.colors:
                continue

            # XXX: same as above
            if self.prev and self.prev.colors.get(color, None) is not None:
                self.colors[color] = self.prev.colors[color]
                self.pixel_state[self.prev.colors[color]] = True

    def get_binary_data(self):
        """
        Return binary data for the char
        """
        result = {"bitmap": [], "screen-ram": 0, "color-ram": 0}

        for row in zip(*[iter(sorted(self.pixels))] * 4):
            char_line = 0
            for idx, pixel in enumerate(row):
                bits = self.colors.get(self.pixels[pixel], (0, 0))
                char_line += bits[0] * 2 ** (7 - idx * 2)
                char_line += bits[1] * 2 ** (6 - idx * 2)
            result["bitmap"].append(char_line)

        colors = dict([(y, x) for x, y in self.colors.items()])
        col1 = colors.get((0, 1), colors.get((0, 0))) * 16
        col2 = colors.get((1, 0), colors.get((0, 0)))

        result["color-ram"] = col1 + col2
        result["screen-ram"] = colors.get((1, 1), colors.get((0, 0)))

        return result


class MultiConverter(base.FullScreenImage):
    """
    Convert bitmap graphic in png/gif/probably other formats supported by
    PIL prepared as multicolor image into executable C64 prg file suitable
    to transfer to real thing or run in emulator.
    """
    WIDTH = 160
    HEIGHT = 200
    LOGGER_NAME = "MultiConverter"

    def __init__(self, fname, errors_action="none"):
        """
        Initialization
        """
        super(MultiConverter, self).__init__(fname, errors_action)
        self._scaled = False
        self._save_map = {"prg": self._save_prg,
                          "raw": self._save_raw,
                          "multi": self._save_koala,  # sane default
                          "koala": self._save_koala}

    def _load(self):
        """
        Load source image and store it under _src_image attribute.
        Shrink it if needed to 160x200 pixels.
        """
        if super(MultiConverter, self)._load():
            if self._src_image.size == (320, 200):
                self._src_image = self._src_image.resize((160, 200))
                self._scaled = True
            return True
        return False

    def _check_dimensions(self):
        """
        Check for image dimensions. If different from 320x200 or 160x200
        return False
        """
        result = super(MultiConverter, self)._check_dimensions()

        width, height = self._src_image.size
        if width == MultiConverter.WIDTH and height == MultiConverter.HEIGHT:
            return True

        if not result:
            self.log.error("Wrong picture dimensions: %dx%d", width, height)
        return result

    def _get_displayer(self):
        """
        Get displayer for multicolor picture (based on kickassembler example)
        """
        border = self._get_border()
        background = self._get_background()

        return bytearray([0x01, 0x08, 0x0b, 0x08, 0x0a, 0x00, 0x9e,
                          0x32, 0x30, 0x36, 0x34,
                          0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0xa9,
                          0x38, 0x8d, 0x18, 0xd0, 0xa9, 0xd8, 0x8d, 0x16,
                          0xd0, 0xa9, 0x3b, 0x8d, 0x11, 0xd0, 0xa9, border,
                          0x8d, 0x20, 0xd0, 0xa9, background, 0x8d, 0x21,
                          0xd0, 0xa2, 0x00, 0xbd, 0x00, 0x1c, 0x9d, 0x00,
                          0xd8, 0xbd, 0x00, 0x1d, 0x9d, 0x00, 0xd9, 0xbd,
                          0x00, 0x1e, 0x9d, 0x00, 0xda, 0xbd, 0x00, 0x1f,
                          0x9d, 0x00, 0xdb, 0xe8, 0xd0, 0xe5, 0x4c, 0x46,
                          0x08])

    def _fill_memory(self):
        """
        Create bitmap, screen-ram, color-ram and error map as a picture if
        needed.
        """
        self.data["bitmap"] = []
        self.data["screen-ram"] = []
        self.data["color-ram"] = []
        self.data["background"] = self._get_background()
        self.data["chars"] = []
        used_color_pairs = {'can_be_chars': True}
        error_list = []
        pixel_state = {(0, 0): self.data['background']}

        for color in self.data['most_freq_colors']:
            if color == self.data['background']:
                continue
            if (0, 1) not in pixel_state:
                pixel_state[(0, 1)] = color
                continue
            if (1, 0) not in pixel_state:
                pixel_state[(1, 0)] = color
            break
        pixel_state = {y: x for x, y in pixel_state.items()}

        # get every char (4x8 pixels) starting from upper left corner
        for chry, chrx in [(chry, chrx)
                           for chry in range(0, self._src_image.size[1], 8)
                           for chrx in range(0, self._src_image.size[0], 4)]:

            box = self._src_image.crop((chrx, chry,
                                        chrx + 4, chry + 8)).convert("RGB")

            char = MultiChar(self.log,
                             pixel_state,
                             self.prev_chars.get((chry, chrx)),
                             self._errors_action == "fix")
            char.process(box, self._palette_map)
            self.chars[(chry, chrx)] = char

            char_data = char.get_binary_data()
            self.data['bitmap'].extend(char_data['bitmap'])
            self.data['screen-ram'].append(char_data['screen-ram'])
            self.data['color-ram'].append(char_data['color-ram'])

            self.prev_chars[(chry, chrx)] = char
            if char.clash:
                error_list.append((chrx, chry))
                self.log.error("Too many colors per block in char %d, %d near"
                               " x=%d, y=%d.",
                               chrx / 8 + 1,
                               chry / 8 + 1,
                               chrx + 4,
                               chry + 4)

            for color, pair in char.colors.items():
                if pair not in used_color_pairs:
                    used_color_pairs[pair] = color
                    continue

                if used_color_pairs[pair] != color and pair != (1, 1):
                    used_color_pairs['can_be_chars'] = False
                    break

        if error_list:
            self._error_image_action(error_list, self._scaled)
            return False

        if used_color_pairs['can_be_chars']:
            self.log.debug(used_color_pairs)
            del used_color_pairs['can_be_chars']
            used_color_pairs = {x: base.COLOR_NAMES[y]
                                for x, y in used_color_pairs.items()}
            self.log.info("Possible char conversion. Pixel pairs and "
                          "corresponding colors: %s", used_color_pairs)
        else:
            self.log.debug("Image cannot be converted to chars")

        self.log.info("Conversion successful.")
        return True

    def _save_prg(self, filename):
        """
        Save executable version of the picture
        """
        file_obj = open(filename, "wb")
        file_obj.write(self._get_displayer())
        file_obj.write(951 * b'\x00')
        file_obj.write(bytearray(self.data["color-ram"]))
        file_obj.write(3096 * b'\x00')
        file_obj.write(bytearray(self.data["screen-ram"]))
        file_obj.write(24 * b'\x00')
        file_obj.write(bytearray(self.data["bitmap"]))
        file_obj.close()
        self.log.info("Saved executable under `%s' file", filename)
        return True

    def _save_koala(self, filename):
        """
        Save as Koala format
        """
        file_obj = open(filename, "wb")
        file_obj.write(bytearray([0x00, 0x60]))
        file_obj.write(bytearray(self.data['bitmap']))
        file_obj.write(bytearray(self.data["color-ram"]))
        file_obj.write(bytearray(self.data["screen-ram"]))
        file_obj.write(bytearray([self.data["background"]]))

        file_obj.close()
        self.log.info("Saved in Koala format under `%s' file", filename)
        return True

    def _save_raw(self, filename):
        """
        Save as raw data
        """

        with open(filename + "_bitmap.raw", "wb") as file_obj:
            file_obj.write(bytearray(self.data['bitmap']))

        with open(filename + "_screen.raw", "wb") as file_obj:
            file_obj.write(bytearray(self.data["screen-ram"]))

        with open(filename + "_color-ram.raw", "wb") as file_obj:
            file_obj.write(bytearray(self.data["color-ram"]))

        with open(filename + "_bg.raw", "wb") as file_obj:
            file_obj.write(bytearray([self.data["background"]]))

        self.log.info("Saved in raw format under `%s_*' files", filename)
        return True



import sys
import logging


class Logger:
    """
    Logger class with output on console only
    """
    def __init__(self, logger_name):
        """
        Initialize named logger
        """
        self._log = logging.getLogger(logger_name)
        self.setup_logger()
        self._log.set_verbose = self.set_verbose

    def __call__(self):
        """
        Calling this object will return configured logging.Logger object with
        additional set_verbose() method.
        """
        return self._log

    def set_verbose(self, verbose_level, quiet_level):
        """
        Change verbosity level. Default level is warning.
        """
        self._log.setLevel(logging.WARNING)

        if quiet_level:
            self._log.setLevel(logging.ERROR)
            if quiet_level > 1:
                self._log.setLevel(logging.CRITICAL)

        if verbose_level:
            self._log.setLevel(logging.INFO)
            if verbose_level > 1:
                self._log.setLevel(logging.DEBUG)

    def setup_logger(self):
        """
        Create setup instance and make output meaningful :)
        """
        if self._log.handlers:
            # need only one handler
            return

        console_handler = logging.StreamHandler(sys.stderr)
        console_handler.set_name("console")
        console_formatter = logging.Formatter("%(levelname)s: %(message)s")
        console_handler.setFormatter(console_formatter)
        self._log.addHandler(console_handler)
        self._log.setLevel(logging.WARNING)



#!/usr/bin/env python
"""
Tests for Image2c64 <https://bitbucket.org/gryf/image2c64>

Author: Roman 'gryf' Dobosz <gryf73@gmail.com>
Date: 2012-11-20
Version: 1.3
Licence: BSD
"""
from tempfile import mkstemp, mkdtemp
from unittest import TestCase, main
import argparse
import os
import shutil

from c64img import cmd_convert
from c64img.base import FullScreenImage


HIRES = os.path.join(os.path.dirname(__file__), "test_images", "hires.png")


class Interceptor:
    """
    Interceptor class for function call detection
    """
    def __init__(self):
        """
        Init.
        """
        self.call = 0
        self.repeat = False

    def __call__(self, dummy1, dummy2=False):
        """
        Call attribute is increased every time, instance is called
        """
        self.call += 1
        return self.repeat


class TestResolvename(TestCase):
    """Test resolve name function"""

    def setUp(self):
        if not hasattr(self, "_path"):
            self._path = os.path.abspath(os.curdir)

        if not hasattr(self, "_args_mock_cls"):
            class Obj:
                """
                Mock args class
                """
                def __init__(self):
                    """
                    Initialization
                    """
                    self.filename = "foo"
                    self.format = "hires"
                    self.output = None
                    self.executable = False
                    self.raw = False
            self._args_mock_cls = Obj

    def tearDown(self):
        os.chdir(self._path)

    def test_resolve_name(self):
        """
        Test resolve_name function
        """
        args = self._args_mock_cls()
        args.filename = ["foo"]
        args.format = "hires"
        args.output = None
        args.executable = False
        self.assertEqual(cmd_convert.resolve_name(args, args.filename[0]),
                         ("foo.prg", "hires"))

        args.output = "bar"
        self.assertEqual(cmd_convert.resolve_name(args, args.filename[0]),
                         ("bar", "hires"))

        args.executable = True
        self.assertEqual(cmd_convert.resolve_name(args, args.filename[0]),
                         ("bar.prg", "prg"))

        args.executable = False
        args.raw = True
        args.output = "bar.prg"
        self.assertEqual(cmd_convert.resolve_name(args, args.filename[0]),
                         ("bar", "raw"))

        args.output = "foo.hires"
        self.assertEqual(cmd_convert.resolve_name(args, args.filename[0]),
                         ("foo", "raw"))

        args.output = "foo.hires.png"
        self.assertEqual(cmd_convert.resolve_name(args, args.filename[0]),
                         ("foo.hires", "raw"))

        dname = mkdtemp()
        os.chdir(dname)

        args.output = "outdir"
        args.filename = ["foo.png", "bar,png"]
        self.assertEqual(cmd_convert.resolve_name(args, args.filename[0]),
                         ("outdir/foo", "raw"))

        fdesc, fname = mkstemp(suffix=".prg", dir=args.output)
        os.close(fdesc)

        args.output = fname
        self.assertRaises(IOError, cmd_convert.resolve_name, args,
                          args.filename[0])

        os.chdir(self._path)
        shutil.rmtree(dname)


class TestMisc(TestCase):
    """
    Test helper functions
    """

    def test_convert(self):
        """
        Test convert function
        """

        class Mock:
            def __init__(self):
                self.filename = []
                self.errors = ['show', 'save', 'none'][2]
                self.border = None
                self.background = None
                self.verbose = 0
                self.quiet = 0
                self.output = None
                self.format = None
                self.executable = False
                self.raw = False

        def mock_set_verbose(verbose, quiet):
            ConvClass.V = verbose
            ConvClass.Q = quiet

        class ConvClass(FullScreenImage):
            BG = None
            BORDER = None
            LOG = None
            V = 0
            Q = 0
            ERRORS = "none"
            SELF = []

            def __init__(self, fname, errors_action="none"):
                super(ConvClass, self).__init__(fname, errors_action)
                ConvClass.ERRORS = errors_action
                self.log = Mock()
                self.log.set_verbose = mock_set_verbose
                ConvClass.SELF.append(self)

            def save(self, fname, format_):
                return

            def set_border_color(self, color):
                ConvClass.BORDER = color

            def set_bg_color(self, color):
                ConvClass.BG = color

        arg_mock = Mock()
        cmd_convert.convert(arg_mock, ConvClass)
        self.assertEqual(ConvClass.BG, None)
        self.assertEqual(ConvClass.BORDER, None)
        self.assertEqual(ConvClass.V, 0)
        self.assertEqual(ConvClass.Q, 0)
        self.assertEqual(ConvClass.ERRORS, 'none')
        self.assertEqual(len(ConvClass.SELF), 0)

        arg_mock = Mock()
        arg_mock.filename = ["foo", "bar"]
        arg_mock.errors = "show"
        cmd_convert.convert(arg_mock, ConvClass)
        self.assertEqual(ConvClass.ERRORS, 'show')
        self.assertEqual(len(ConvClass.SELF), 2)
        self.assertEqual(ConvClass.SELF[0].prev_chars, {})
        self.assertEqual(ConvClass.SELF[1].prev_chars, ConvClass.SELF[0].chars)

        arg_mock = Mock()
        arg_mock.filename = ["foo", "bar", "baz"]
        arg_mock.border = 2  # red
        arg_mock.background = 8
        cmd_convert.convert(arg_mock, ConvClass)
        self.assertEqual(ConvClass.BORDER, 2)
        self.assertEqual(ConvClass.BG, 8)
        self.assertEqual(len(ConvClass.SELF), 5)
        self.assertEqual(ConvClass.SELF[2].prev_chars, {})
        self.assertEqual(ConvClass.SELF[3].prev_chars, ConvClass.SELF[2].chars)
        self.assertEqual(ConvClass.SELF[4].prev_chars, ConvClass.SELF[3].chars)

        arg_mock = Mock()
        arg_mock.filename = ["foo"]
        arg_mock.verbose = 2
        arg_mock.quiet = 2
        cmd_convert.convert(arg_mock, ConvClass)
        self.assertEqual(ConvClass.V, 2)
        self.assertEqual(ConvClass.Q, 2)
        self.assertEqual(ConvClass.ERRORS, 'none')
        self.assertEqual(len(ConvClass.SELF), 6)


class TestCmd(TestCase):
    """Test cmd_convert functions"""

    def setUp(self):
        """Setup"""
        class AnyObj:
            """General mock"""
            pass

        class ArgParseMock:
            """ArgumentParser mock"""

            def __init__(self, *args, **kwargs):
                pass

            def add_argument(self, *args, **kwargs):
                pass

            def add_mutually_exclusive_group(self, *args, **kwargs):
                return ArgParseMock()

            def parse_args(self):
                retobj = AnyObj()
                retobj.format = "hires"
                return retobj

        self._argparse = argparse.ArgumentParser
        argparse.ArgumentParser = ArgParseMock
        self._convert = cmd_convert.convert
        cmd_convert.convert = lambda x, y: 0

    def tearDown(self):
        """Teardown"""
        argparse.ArgumentParser = self._argparse
        cmd_convert.convert = self._convert

    def test_image2c64(self):
        """
        Test image2c64 cmd_convert
        """
        self.assertRaises(SystemExit, cmd_convert.image2c64)


if __name__ == "__main__":
    main()



"""
Fullscreen high resolution image abstraction.

Assumptions:
    - Input image 320x200 pixels.
    - Maximum 2 colors per char may be used.
"""
from c64img import base


class HiresChar(base.Char):
    """
    Hires char implementation
    """

    def get_binary_data(self):
        """
        Return binary data for the char
        """
        result = {"bitmap": [], "screen-ram": 0}

        pixel_keys = sorted(self.pixels.keys())

        for key_row in [pixel_keys[index:index+8]
                        for index in range(0, len(pixel_keys), 8)]:
            char_line = 0
            for key in key_row:
                bit_ = self.colors.get(self.pixels[key], 0)
                char_line += bit_ * 2 ** (7 - key[1])
            result['bitmap'].append(char_line)

        colors = dict([(y, x) for x, y in self.colors.items()])
        if 0 in colors:
            result['screen-ram'] = colors[0]

        if 1 in colors:
            result['screen-ram'] += colors[1] * 16
        return result


class HiresConverter(base.FullScreenImage):
    """
    Convert bitmap graphic in png/gif/probably other formats supported by
    PIL into executable C64 prg file suitable to transfer to real thing or
    run in emulator.
    """
    LOGGER_NAME = "HiresConverter"

    def __init__(self, fname, errors_action="none"):
        """
        Initialization
        """
        super(HiresConverter, self).__init__(fname, errors_action)
        self._save_map = {"prg": self._save_prg,
                          "raw": self._save_raw,
                          "hires": self._save_ash,  # make sane default
                          "art-studio-hires": self._save_ash}

    def _get_displayer(self):
        """
        Get displayer for hires picture
        """
        return bytearray([0x01, 0x08, 0x0b, 0x08, 0x0a, 0x00, 0x9e, 0x32,
                          0x30, 0x36, 0x34, 0x00, 0x00, 0x00, 0x00, 0x00,
                          0x00, 0x78, 0xa9, self._get_border(), 0x8d, 0x20,
                          0xd0, 0xa9, 0x00, 0x8d, 0x21, 0xd0, 0xa9, 0xbb,
                          0x8d, 0x11, 0xd0, 0xa9, 0x3c, 0x8d, 0x18, 0xd0,
                          0x4c, 0x25, 0x08])

    def _fill_memory(self):
        """
        Create bitmap/screen and error map as a picture if needed.
        """
        self.data["bitmap"] = []
        self.data["screen-ram"] = []
        error_list = []

        for chry, chrx in [(chry, chrx)
                           for chry in range(0, self._src_image.size[1], 8)
                           for chrx in range(0, self._src_image.size[0], 8)]:

            box = self._src_image.crop((chrx, chry,
                                        chrx + 8, chry + 8)).convert("RGB")

            char = HiresChar(self.log,
                             self.prev_chars.get((chry, chrx)),
                             self._errors_action == "fix")
            char.process(box, self._palette_map)
            self.chars[(chry, chrx)] = char

            char_data = char.get_binary_data()
            self.data['bitmap'].extend(char_data['bitmap'])
            self.data['screen-ram'].append(char_data['screen-ram'])

            if char.clash:
                error_list.append((chrx, chry))
                self.log.error("Too many colors per block in char %d, %d near"
                               " x=%d, y=%d.",
                               chrx / 8 + 1,
                               chry / 8 + 1,
                               chrx + 4,
                               chry + 4)

        if error_list:
            self._error_image_action(error_list)
            return False

        self.log.info("Conversion successful.")
        return True

    def _save_prg(self, filename):
        """
        Save executable version of the picture
        """
        file_obj = open(filename, "wb")
        file_obj.write(self._get_displayer())
        file_obj.write(984 * b'\x00')
        file_obj.write(bytearray(self.data["screen-ram"]))
        file_obj.write(4120 * b'\x00')
        file_obj.write(bytearray(self.data["bitmap"]))
        file_obj.close()
        self.log.info("Saved executable under `%s' file", filename)
        return True

    def _save_ash(self, filename):
        """
        Save as Art Studio hires
        """
        file_obj = open(filename, "wb")
        file_obj.write(bytearray([0x00, 0x20]))
        file_obj.write(bytearray(self.data['bitmap']))
        file_obj.write(bytearray(self.data["screen-ram"]))
        file_obj.write(bytearray([self._get_border()]))
        file_obj.write(6 * b'\x00')
        file_obj.close()
        self.log.info("Saved in Art Studio Hires format under `%s' file",
                      filename)
        return True

    def _save_raw(self, filename):
        """
        Save raw data
        """
        with open(filename + "_screen.raw", "wb") as file_obj:
            file_obj.write(bytearray(self.data["screen-ram"]))

        with open(filename + "_bitmap.raw", "wb") as file_obj:
            file_obj.write(bytearray(self.data["bitmap"]))

        self.log.info("Saved raw data under `%s_*' files", filename)
        return True

    def _check_dimensions(self):
        """
        Check for image dimensions. Same as in superclass, needed for feedback
        only.
        """
        result = super(HiresConverter, self)._check_dimensions()
        width, height = self._src_image.size

        if not result:
            self.log.error("Wrong picture dimensions: %dx%d", width, height)

        return result



"""
C64lib module holds all information needed to provide abstraction to low level
graphic related operation and conbertion in Python.
"""
import os
from collections import Counter
import logging

from PIL import Image
from PIL.ImageDraw import Draw

from c64img.path import get_modified_fname
from c64img.logger import Logger

# Palettes are organized with original C64 order:
# black, white, red, magenta, purple, green, dark blue, yellow,
# orange, brown, pink, dark gray, gray, light green, light blue, light gray

(BLACK, WHITE, RED, MAGENTA, PURPLE, GREEN, DARK_BLUE, YELLOW,
 ORANGE, BROWN, PINK, DARK_GRAY, GRAY, LIGHT_GREEN, LIGHT_BLUE,
 LIGHT_GRAY) = range(16)

PALETTES = {'Vice': ((0x00, 0x00, 0x00), (0xFF, 0xFF, 0xFF),
                     (0x9a, 0x53, 0x48), (0x8c, 0xc9, 0xd1),
                     (0x9b, 0x5a, 0xbb), (0x7b, 0xb7, 0x54),
                     (0x51, 0x42, 0xb1), (0xd8, 0xe2, 0x84),
                     (0xa0, 0x72, 0x35), (0x70, 0x5b, 0x00),
                     (0xc6, 0x89, 0x80), (0x69, 0x69, 0x69),
                     (0x92, 0x92, 0x92), (0xb9, 0xee, 0x99),
                     (0x8e, 0x82, 0xe1), (0xb9, 0xb9, 0xb9)),
            'Timanthes': ((0, 0, 0), (213, 213, 213),
                          (114, 53, 44), (101, 159, 166),
                          (115, 58, 145), (86, 141, 53),
                          (46, 35, 125), (174, 183, 94),
                          (119, 79, 30), (75, 60, 0),
                          (156, 99, 90), (71, 71, 71),
                          (107, 107, 107), (143, 194, 113),
                          (103, 93, 182), (143, 143, 143)),
            'Unknown': ((0x00, 0x00, 0x00), (0xff, 0xff, 0xff),
                        (0x88, 0x00, 0x00), (0xaa, 0xff, 0xee),
                        (0xcc, 0x44, 0xcc), (0x00, 0xcc, 0x55),
                        (0x00, 0x00, 0xaa), (0xee, 0xee, 0x77),
                        (0xdd, 0x88, 0x55), (0x66, 0x44, 0x00),
                        (0xff, 0x77, 0x77), (0x33, 0x33, 0x33),
                        (0x77, 0x77, 0x77), (0xaa, 0xff, 0x66),
                        (0x00, 0x88, 0xff), (0xbb, 0xbb, 0xbb)),
            'Pepto': ((0x00, 0x00, 0x00), (0xFF, 0xFF, 0xFF),
                      (0x68, 0x37, 0x2B), (0x70, 0xA4, 0xB2),
                      (0x6F, 0x3D, 0x86), (0x58, 0x8D, 0x43),
                      (0x35, 0x28, 0x79), (0xB8, 0xC7, 0x6F),
                      (0x6F, 0x4F, 0x25), (0x43, 0x39, 0x00),
                      (0x9A, 0x67, 0x59), (0x44, 0x44, 0x44),
                      (0x6C, 0x6C, 0x6C), (0x9A, 0xD2, 0x84),
                      (0x6C, 0x5E, 0xB5), (0x95, 0x95, 0x95))}
COLOR_SUBS = {BLACK: [DARK_GRAY, BROWN, DARK_BLUE],
              WHITE: [LIGHT_GRAY, LIGHT_GRAY],
              RED: [DARK_GRAY, GRAY, PURPLE],
              MAGENTA: [LIGHT_GRAY, GRAY, LIGHT_BLUE],
              PURPLE: [GRAY, DARK_GRAY, RED, ORANGE],
              GREEN: [GRAY, LIGHT_GRAY, YELLOW, LIGHT_GREEN],
              DARK_BLUE: [DARK_GRAY, BROWN, LIGHT_BLUE, BLACK],
              YELLOW: [LIGHT_GRAY, LIGHT_GREEN, WHITE],
              ORANGE: [GRAY, DARK_GRAY, PURPLE, BROWN],
              BROWN: [DARK_GRAY, DARK_BLUE, RED, BLACK],
              PINK: [GRAY, LIGHT_GRAY, GREEN, RED],
              DARK_GRAY: [GRAY, BROWN, DARK_BLUE, BLACK],
              GRAY: [LIGHT_GRAY, GREEN, LIGHT_BLUE, ORANGE, PURPLE, PINK],
              LIGHT_GREEN: [LIGHT_GRAY, YELLOW, WHITE],
              LIGHT_BLUE: [GRAY, LIGHT_GRAY, MAGENTA],
              LIGHT_GRAY: [LIGHT_GREEN, YELLOW, GRAY, WHITE]}
COLOR_NAMES = dict(enumerate(["Black", "White", "Red", "Magenta", "Purple",
                              "Green", "Dark blue", "Yellow", "Orange",
                              "Brown", "Pink", "Dark gray", "Gray",
                              "Light green", "Light blue", "Light gray"]))


class Char:
    """
    Char implementation
    """

    def __init__(self, log, prev=None, fix_clash=False):
        """
        Init. prev is the Char object which represents the same character in
        previous picture
        """
        self.background = None
        self.clash = False
        self.colors = {}
        self.max_colors = 2
        self.pixels = {}
        self.pixel_state = {0: False, 1: False}
        self.prev = prev
        self.log = log
        self._fix_clash = fix_clash

    def _analyze_color_map(self):
        """
        Check for the optimal color placement in char. This method may be run
        only on not clashed chars.
        """

        if self._check_clash():
            if not self._fix_clash:
                return
            self._fix_color_clash()

        if self.prev and self.prev.clash:
            return

        colors = Counter(self.pixels.values())
        self._compare_colors(colors)

    def get_binary_data(self):
        """
        Return binary data for the char
        """
        raise NotImplementedError

    def process(self, box, palette_map):
        """
        Store pixel/color information
        """
        for chry, chrx in sorted([(y, x)
                                  for x in range(box.size[0])
                                  for y in range(box.size[1])]):
            self.pixels[(chry, chrx)] = palette_map[box.getpixel((chrx, chry))]

        self._analyze_color_map()

    def _fix_color_clash(self):
        """
        Try to fix the color clashes by assigning excessed colors to existing
        in the char
        """
        colors = Counter(self.pixels.values())
        base = colors.most_common(self.max_colors)
        remapped_colors = {}
        index = self.max_colors

        if self.background is not None and self.background not in dict(base):
            index -= 1
            base = colors.most_common(index)
            base += [(self.background, 0)]

        for col, dummy in colors.most_common()[index:]:
            sub = _get_the_substitute(col, base)
            if sub is not None:
                self.log.debug("Using color '%s' instead '%s'",
                               COLOR_NAMES[sub],
                               COLOR_NAMES[col])
                remapped_colors[col] = sub
            elif self.background is not None:
                self.log.warning("Cannot remap color; using background - '%s'",
                                 COLOR_NAMES[self.background])
                remapped_colors[col] = self.background
            else:
                self.log.warning("Cannot remap color; using first - '%s'",
                                 COLOR_NAMES[base[0][0]])
                remapped_colors[col] = base[0][0]

        for coord, color in self.pixels.items():
            if color not in remapped_colors:
                continue
            self.pixels[coord] = remapped_colors[color]

        # Clashes are fixed (probably)
        self.clash = False

    def _check_clash(self):
        """
        Check color clash. max_colors is the maximum colors per char object
        """
        colors = Counter(self.pixels.values())

        if len(colors) > self.max_colors:
            self.clash = True

        if self.background is None:
            return self.clash

        if len(colors) == self.max_colors and self.background not in colors:
            self.clash = True

        return self.clash

    def _compare_colors_with_prev_char(self, colors, repeat=False):
        """
        Make a color map to the pixels comparing to the previous data
        """
        needs_repeat = False
        for color in sorted(colors):
            if color == self.background:
                continue
            if repeat:
                if color in self.colors:
                    continue

                for pair, taken in self.pixel_state.items():
                    if not taken:
                        self.pixel_state[pair] = True
                        self.colors[color] = pair
                        break
            elif self.prev and self.prev.colors.get(color, None) is not None:
                self.colors[color] = self.prev.colors[color]
                self.pixel_state[self.prev.colors[color]] = True
            else:
                needs_repeat = True

        return needs_repeat

    def _compare_colors(self, colors):
        """
        Make a color map to the pixels
        """
        if self._compare_colors_with_prev_char(colors):
            self._compare_colors_with_prev_char(colors, True)


class FullScreenImage:
    """
    Class represents full-screen image in unspecified C64 format
    """
    WIDTH = 320
    HEIGHT = 200
    LOGGER_NAME = 'PictureConverter'

    def __init__(self, fname, errors_action="none"):
        """
        Initialization. Fname is a full/relative path for a hires picture.
        errors_action can be one of "none", "save" and "show". In case of
        color clashes, it will do nothing, save picture with "_error" suffix
        or show it respectively.
        """
        self._errors_action = errors_action
        self._fname = fname
        self._palette_map = None
        self._save_map = {}
        self._src_image = None
        self.chars = {}
        self.data = {}
        self.log = Logger(self.LOGGER_NAME)()
        self.prev_chars = {}

    def save(self, filename, format_=None):
        """
        Save picture in one of the formats or as an executable prg.
        """
        if not self.data.get('bitmap'):
            if not self.convert():
                return 1

        if os.path.exists(filename):
            self.log.warning("File `%s' will be overwritten", filename)

        if self._save_map[format_](filename):
            return 0

    def set_bg_color(self, color):
        """
        Set custom color for background (as index)
        """
        self.log.info("Setting background color to: %d", color)
        self.data['background'] = color

    def set_border_color(self, color):
        """
        Set custom color for border (as index)
        """
        self.log.info("Setting border color to: %d", color)
        self.data['border'] = color

    def convert(self):
        """
        Convert image to binary data
        """
        if not self._load():
            return False

        if not self._check_dimensions():
            return False

        hist = self._src_image.histogram()
        self._colors_check(hist)
        self._find_best_palette_map()
        self._find_most_freq_color(hist)
        return self._fill_memory()

    def _load(self):
        """
        Load src image and store it under _src_image attribute.
        """
        try:
            img = Image.open(self._fname)
            img = img.convert("RGB").convert('P', palette=Image.ADAPTIVE,
                                             dither='none', colors=16)
            self._src_image = img
        except IOError:
            self.log.critical("Cannot open file `%s'. Exiting.", self._fname)
            if self.log.getEffectiveLevel() == logging.DEBUG:
                raise
            return False
        return True

    def _check_dimensions(self):
        """
        Check for image dimensions. If different from 320x200 return False
        """
        width, height = self._src_image.size

        if width == FullScreenImage.WIDTH and height == FullScreenImage.HEIGHT:
            return True

        return False

    def _find_most_freq_color(self, histogram):
        """
        Check for the most frequent color on the picture. Can be used to
        auto detect background/border colors.
        Value remembered in attribute data['most_freq_color'] is an
        index in the C64 palette (NOT the source image palette!).
        """
        pal = self._src_image.getpalette()
        pal = [tuple(pal[index:index + 3]) for index in range(0, len(pal), 3)]
        sorted_hist = sorted([(count, index)
                              for index, count in enumerate(histogram[:16])],
                             reverse=True)
        self.data['most_freq_colors'] = []
        for _, index in sorted_hist:
            try:
                color = self._palette_map[pal[index]]
            except IndexError:
                # image palette have incomplete amount of colors - ignore
                continue
            self.data['most_freq_colors'].append(color)

        self.data['most_freq_color'] = self.data['most_freq_colors'][0]

    def _colors_check(self, histogram):
        """
        Find out how many same colors do we have. Just an information to the
        user.
        """
        no_of_colors = len([col for col in histogram if col != 0])
        if no_of_colors < 2:
            self.log.warn("Picture have %d color(s). Result may be confusing.",
                          no_of_colors)
        else:
            self.log.info("Picture have %d colors", no_of_colors)

        return no_of_colors

    def _get_palette(self):
        """
        Return source image palette as RGB tuples
        """
        pal = self._src_image.getpalette()
        if len(pal) < 48:
            # image does not have full palette - fill gap with 0
            pal.extend([0 for _ in range(48-len(pal))])
        return [(pal[i], pal[i + 1], pal[i + 2]) for i in range(0, 16 * 3, 3)]

    def _fill_memory(self):
        """
        Create bitmap/screen-ram/color-ram colors if needed. Should be
        implemented in concrete implementation.
        """
        raise NotImplementedError

    def _find_best_palette_map(self):
        """
        Try to match source image palette to predefined ones, and return name
        of the best matched palette and color map for current source image.
        """
        palettes_map = {}
        src_pal = self._get_palette()
        nearest_match = -1
        selected_palette = None

        for pal_name, pal in PALETTES.items():
            quality = 0
            palettes_map[pal_name] = {}
            for orig_color in src_pal:
                color2use, delta = _best_color_match(orig_color, pal)
                quality += delta
                palettes_map[pal_name][orig_color] = color2use
            if nearest_match < 0 or nearest_match > quality:
                nearest_match = quality
                selected_palette = pal_name

        if nearest_match == 0:
            self.log.info("Perfect palette match for %s-palette",
                          selected_palette)
        else:
            self.log.info("Quality: %d for %s-palette ", nearest_match,
                          selected_palette)

        self._palette_map = palettes_map[selected_palette]

    def _get_border(self):
        """
        Return border color index
        """
        border = self.data.get('border')
        if border is None:
            border = self.data.get('most_freq_color', 0)

        return border

    def _get_background(self):
        """
        Return background color index
        """
        background = self.data.get("background")
        if background is None:
            background = self.data.get("most_freq_color", 0)

        self.log.debug("Setting color '%s' as background.",
                       COLOR_NAMES[background])
        return background

    def _error_image_action(self, error_list, scaled=False):
        """
        Create image with hints of clashes. error_list contains coordinates of
        characters which encounters clashes. Picture is for overview only,so
        it is always have dimensions 320x200.
        """
        char_x_size = 1
        x_offset = 3
        if self._errors_action == "none":
            return

        image = self._src_image.copy().convert("RGBA")

        # TODO: refactor this crap below
        if scaled:
            image = image.resize((320, 200))

        if image.size[0] == 320:
            x_offset = 7
            if scaled:
                char_x_size = 2

        if image.size[0] == 160:
            char_x_size = 1

        image_map = image.copy()
        drawable = Draw(image_map)

        for chrx, chry in error_list:
            drawable.rectangle((chrx * char_x_size,
                                chry,
                                chrx * char_x_size + x_offset,
                                chry + 7),
                               outline="red")

        image = Image.blend(image, image_map, 0.65)
        del drawable

        if self._errors_action in ('save', 'grafx2'):
            file_obj = open(get_modified_fname(self._fname, 'png', '_error.'),
                            "wb")
            image.save(file_obj, "png")
            file_obj.close()
            if self._errors_action == 'grafx2':
                os.system("grafx2 %s %s" %
                          (self._fname, get_modified_fname(self._fname, 'png',
                                                           '_error.')))
        else:
            clashes = image.resize((640, 400))
            clashes.show()


def _get_the_substitute(color, base):
    """Return best match for provided color out of base and background"""
    sub_color = None
    index = 16
    for col, dummy in base:
        if col in COLOR_SUBS[color]:
            if COLOR_SUBS[color].index(col) < index:
                index = COLOR_SUBS[color].index(col)
                sub_color = col

    return sub_color


def _best_color_match(orig_color, colors):
    """
    Match provided color for closed match in colors list, and return it's
    index and delta (which indicates similarity)
    """
    delta = 0
    src_r, src_g, src_b = orig_color

    rgb_diff = 195075  # arbitrary number of maximum difference
    for idx, (pal_r, pal_g, pal_b) in enumerate(colors):
        partial_delta = (src_r - pal_r) ** 2 + \
                        (src_g - pal_g) ** 2 + (src_b - pal_b) ** 2

        if partial_delta == 0:  # bingo
            delta = partial_delta
            color2use = idx
            break

        if rgb_diff > partial_delta:
            rgb_diff = partial_delta
            delta = rgb_diff
            color2use = idx

    return color2use, delta
